#include <stdio.h>
#include <string.h>
#include <io.h>
#include <fcntl.h>
#include <sys/stat.h>
#include "../version.h"
#include "driver.h"

char * camlrun = "camlrun.exe";
char * stdlib;
char * includes = "";
char * compopt = "";
char * compfiles = "";
char * linkfiles = "";
char * output = "camltop.exe";
int custom = 0;
char * ccfiles = "";
char * ccopt = "";
char * perv = "";

char syst_perv[] =
   "baltree bool char eq exc fchar filename float format fstring fvect "
   "gc genlex hashtbl int io iparsing lexing list map obj pair parsing "
   "printexc printf queue random ref set sort stack stream string "
   "toplevel vect";

main(argc, argv)
     int argc;
     char ** argv;
{
  int i;
  char * a;
  extern char * getenv();
  char * cmd;
  int status;
  
  stdlib = getenv("CAMLLIB");
  if (stdlib == NULL) {
    fprintf(stderr, "Variable CAMLLIB is undefined.\n");
    exit(2);
  }

  for (i = 1; i < argc; i++) {
    a = argv[i];
    if (suffix(a, ".zo")) {
      reformat(&linkfiles, "%s %s", linkfiles, a);
      reformat(&perv, "%s %s", perv, modulename(a));
    } else
    if (suffix(a, ".zi")) {
      reformat(&perv, "%s %s", perv, modulename(a));
    } else
    if (eq(a, "-I") || eq(a, "-include")) {
      reformat(&includes, "%s -I %s", includes, argv[++i]);
    } else
    if (eq(a, "-o")) {
      output = argv[++i];
    } else
    if (eq(a, "-stdlib")) {
      stdlib = argv[++i];
    } else
    if (eq(a, "-custom")) {
      custom = 1;
    } else
    if (suffix(a, ".c")) {
      format(&cmd, "-c -I%s %s %s", stdlib, ccopt, a);
      status = execute("gcc", cmd);
      if (status != 0) exit(status);
      free(cmd);
      a[strlen(a) - 2] = 0;     /* remove .c extension */
      reformat(&ccfiles, "%s %s.o", ccfiles, a);
    } else
    if (suffix(a, ".o") || suffix(a, ".a") || prefix(a, "-l")) {
      reformat(&ccfiles, "%s %s", ccfiles, a);
    } else
    if (eq(a, "-ccopt")) {
      reformat(&ccopt, "%s %s", ccopt, argv[++i]);
    } else
    if (prefix(a, "-")) {
      fprintf(stderr, "Unknown option \"%s\", ignored\n", a);
    } else {
      fprintf(stderr, "I don't know what to do with file \"%s\", ignored\n",
              a);
    }
  }
  {
  static char require[] = "camlXXXX.req";
  static char temp_out[] = "camlXXXX.byt";
  tempname(require);
  tempname(temp_out);
  format(&cmd, "%s\\provide -stdlib %s %s %s %s",
         stdlib, stdlib, includes, syst_perv, perv);
  status = execute_with_output_to(camlrun, cmd, require);
  if (status != 0) exit(status);
  free(cmd);

  if (!custom) {
    format(&cmd,
           "%s\\camllink -stdlib %s -require %s -exec %s "
           "-g %s stdlib.zo %s toplib.zo",
           stdlib, stdlib, require, temp_out,
           includes, linkfiles);
    status = execute(camlrun, cmd);
    unlink(require);
    free(cmd);
    if (status != 0) exit(status);
    format(&cmd,
           "%s\\expunge %s %s sys %s %s",
           stdlib, temp_out, output, syst_perv, perv);
    status = execute(camlrun, cmd);
    unlink(temp_out);
    free(cmd);
    exit(status);

  } else {

    static char expunged_out[] = "expnXXXX.byt";
    static char prim_file[] = "primXXXX.c";
    char * gcc_out, * exefile;
    tempname(expunged_out);
    tempname(prim_file);
    gcc_out = change_suffix(output, ".cff");
    exefile = change_suffix(gcc_out, ".exe");
    format(&cmd,
           "%s\\camllink -stdlib %s -custom %s -require %s -exec %s "
           "-g %s stdlib.zo %s toplib.zo",
           stdlib, stdlib, prim_file, require, temp_out,
           includes, linkfiles);
    status = execute(camlrun, cmd);
    if (status != 0) goto end;
    free(cmd);
    format(&cmd,
           "%s\\expunge %s %s sys %s %s",
           stdlib, temp_out, expunged_out, syst_perv, perv);
    status = execute(camlrun, cmd);
    if (status != 0) goto end;
    free(cmd);
    format(&cmd,
           "-I%s -o %s %s %s -L%s %s -lcaml -lgrx -lpc -lm",
           stdlib, gcc_out, ccopt, prim_file, stdlib, ccfiles);
    status = execute("gcc", cmd);
    free(cmd);
    if (status != 0) goto end;
    format(&cmd, "-s %s\\go32.exe %s", stdlib, gcc_out);
    status = execute("coff2exe", cmd);
    free(cmd);
    if (status != 0) goto end;
    status = append(expunged_out, exefile);
    free(cmd);
  end:
    unlink(require);
    unlink(temp_out);
    unlink(expunged_out);
    unlink(prim_file);
    unlink(gcc_out);
    if (status != 0) unlink(exefile);
    exit(status);
  }
  }
  return 0;
}
